"""Queue for bytes."""
from __future__ import annotations

import threading


class ByteQueue:
    """FIFO class for queuing and retrieving bytes."""

    def __init__(self) -> None:
        """Initialize the queue."""
        self._buffer = bytearray()
        self._buffer_lock = threading.Condition()

    def append(self, data: bytes):
        """Add bytes to the end of the queue.

        Args:
            data: bytes to add

        """
        with self._buffer_lock:
            self._buffer.extend(data)
            self._buffer_lock.notify_all()

    def pop(self, size: int = 1) -> bytes:
        """Remove and return bytes from the beginning of queue.

        Args:
            size: number of bytes to remove

        Returns:
            removed bytes

        """
        with self._buffer_lock:
            data = self._buffer[:size]
            del self._buffer[:size]
            return data

    def pop_byte(self) -> int:
        """Remove and return single byte from the beginning of queue.

        Returns:
            removed byte

        """
        with self._buffer_lock:
            data = self._buffer[0]
            del self._buffer[0]
            return data

    def peek(self, size: int = 1) -> bytes:
        """Get bytes from beginning of the queue without removing them.

        Args:
            size: number of bytes to peek

        Returns:
            peek bytes

        """
        return self._buffer[:size]

    def peek_byte(self, position: int = 0) -> int:
        """Get single byte in the buffer without removing.

        Args:
            position: byte position to peek at

        Returns:
            peek bytes

        """
        return self._buffer[position]

    def clear(self):
        """Clear the bytes in the queue."""
        with self._buffer_lock:
            self._buffer.clear()

    def __len__(self) -> int:
        """Get the length of the queue.

        Returns:
            queue length

        """
        return len(self._buffer)

    def wait_for(self, size: int = 1, peek: bool = False) -> bytes:
        """Wait until the requested number of bytes is available in the receive queue.

        Args:
            size: number of bytes
            peek: only look, don't remove the item from the queue.

        Returns:
            Found bytes

        """
        def min_size() -> bool:
            return len(self._buffer) >= size

        if len(self._buffer) < size:
            with self._buffer_lock:
                self._buffer_lock.wait_for(min_size)

        if peek:
            return self.peek(size)

        return self.pop(size)

    def wait_for_byte(self, peek: bool = False) -> int:
        """Wait until one byte is available in the receive queue.

        Args:
            peek: only look, don't remove the item from the queue.

        Returns:
            Found byte

        """
        return self.wait_for(peek=peek)[0]



"""Class for stream 05 function 06."""

from secsgem.secs.data_items import ALCD, ALID, ALTX
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS05F06(SecsStreamFunction):
    """list alarms - data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F06
        [
            {
                ALCD: B[1]
                ALID: U1/U2/U4/U8/I1/I2/I4/I8
                ALTX: A[120]
            }
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F06([
        ...     {"ALCD": secsgem.secs.data_items.ALCD.PERSONAL_SAFETY |
        ...              secsgem.secs.data_items.ALCD.ALARM_SET,
        ...      "ALID": 100,
        ...      "ALTX": "text"}])
        S5F6
          <L [1]
            <L [3]
              <B 0x81>
              <U1 100 >
              <A "text">
            >
          > .

    Data Items:
        - :class:`ALCD <secsgem.secs.data_items.ALCD>`
        - :class:`ALID <secsgem.secs.data_items.ALID>`
        - :class:`ALTX <secsgem.secs.data_items.ALTX>`

    """

    _stream = 5
    _function = 6

    _data_format = [
        [
            ALCD,
            ALID,
            ALTX
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""Class for stream 12 function 05."""

from secsgem.secs.data_items import IDTYP, MAPFT, MID, MLCL
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS12F05(SecsStreamFunction):
    """map transmit inquire.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F05
        {
            MID: A/B[80]
            IDTYP: B[1]
            MAPFT: B[1]
            MLCL: U1/U2/U4/U8
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F05({
        ...     "MID": "materialID",
        ...     "IDTYP": secsgem.secs.data_items.IDTYP.WAFER,
        ...     "MAPFT": secsgem.secs.data_items.MAPFT.ARRAY,
        ...     "MLCL": 0})
        S12F5 W
          <L [4]
            <A "materialID">
            <B 0x0>
            <B 0x1>
            <U1 0 >
          > .

    Data Items:
        - :class:`MID <secsgem.secs.data_items.MID>`
        - :class:`IDTYP <secsgem.secs.data_items.IDTYP>`
        - :class:`MAPFT <secsgem.secs.data_items.MAPFT>`
        - :class:`MLCL <secsgem.secs.data_items.MLCL>`

    """

    _stream = 12
    _function = 5

    _data_format = [
        MID,
        IDTYP,
        MAPFT,
        MLCL
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Class for stream 02 function 49."""

from secsgem.secs.data_items import CEPVAL, CPNAME, DATAID, OBJSPEC, RCMD
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F49(SecsStreamFunction):
    """Enhanced remote command.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F49
        {
            DATAID: U1/U2/U4/U8/I1/I2/I4/I8/A
            OBJSPEC: A
            RCMD: U1/I1/A
            PARAMS: [
                {
                    CPNAME: U1/U2/U4/U8/I1/I2/I4/I8/A
                    CEPVAL: L/BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A/B
                }
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F49({
        ...     "DATAID": 1,
        ...     "OBJSPEC": 'test',
        ...     "RCMD": "COMMAND1",
        ...     "PARAMS": [{
        ...         "CPNAME": "CPARAM",
        ...         "CEPVAL": "VALUE"
        ...     }]})
        S2F49
          <L [4]
            <U1 1 >
            <A "test">
            <A "COMMAND1">
            <L [1]
              <L [2]
                <A "CPARAM">
                <A "VALUE">
              >
            >
          > .

    Data Items:
        - :class:`DATAID <secsgem.secs.data_items.DATAID>`
        - :class:`OBJSPEC <secsgem.secs.data_items.OBJSPEC>`
        - :class:`RCMD <secsgem.secs.data_items.RCMD>`
        - :class:`CPNAME <secsgem.secs.data_items.CPNAME>`
        - :class:`CEPVAL <secsgem.secs.data_items.CEPVAL>`

    """

    _stream = 2
    _function = 49

    _data_format = [
        DATAID,
        OBJSPEC,
        RCMD,
        [
            [
                "PARAMS",
                CPNAME,
                CEPVAL
            ]
        ]
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""State machine for control state."""
from __future__ import annotations

import enum

import secsgem.common


class ControlState(enum.Enum):
    """States for control state machine."""

    INIT = 0
    CONTROL = 1
    OFFLINE = 2
    EQUIPMENT_OFFLINE = 3
    ATTEMPT_ONLINE = 4
    HOST_OFFLINE = 5
    ONLINE = 6
    ONLINE_LOCAL = 7
    ONLINE_REMOTE = 8


class ControlStateMachine(secsgem.common.StateMachine):  # pylint: disable=too-many-instance-attributes
    """Control state machine."""

    def __init__(
        self,
        initial_control_state: str = "ATTEMPT_ONLINE",
        initial_online_control_state: str = "REMOTE",
    ) -> None:
        """Initialize state machine."""
        super().__init__()

        self._initial_control_states = ["EQUIPMENT_OFFLINE", "ATTEMPT_ONLINE", "HOST_OFFLINE", "ONLINE"]
        self._initial_control_state = initial_control_state

        self._online_control_states = ["LOCAL", "REMOTE"]
        self._online_control_state = initial_online_control_state

        self.init = secsgem.common.State(
            ControlState.INIT,
            "INIT",
            initial=True)
        self.control = secsgem.common.State(
            ControlState.CONTROL,
            "CONTROL")
        self.offline = secsgem.common.State(
            ControlState.OFFLINE,
            "OFFLINE")
        self.equipment_offline = secsgem.common.State(
            ControlState.EQUIPMENT_OFFLINE,
            "EQUIPMENT_OFFLINE")
        self.attempt_online = secsgem.common.State(
            ControlState.ATTEMPT_ONLINE,
            "ATTEMPT_ONLINE")
        self.host_offline = secsgem.common.State(
            ControlState.HOST_OFFLINE,
            "HOST_OFFLINE")
        self.online = secsgem.common.State(
            ControlState.ONLINE,
            "ONLINE")
        self.online_local = secsgem.common.State(
            ControlState.ONLINE_LOCAL,
            "ONLINE_LOCAL")
        self.online_remote = secsgem.common.State(
            ControlState.ONLINE_REMOTE,
            "ONLINE_REMOTE")

        # transition 1
        self._current_state: secsgem.common.State = self.init

        self._transitions: list[secsgem.common.Transition] = [
                secsgem.common.Transition(
                    "start",
                    self.init,
                    self.control),  # 1
                secsgem.common.Transition(
                    "initial_offline",
                    self.control,
                    self.offline),  # 1
                secsgem.common.Transition(
                    "initial_equipment_offline",
                    self.offline,
                    self.equipment_offline),  # 2
                secsgem.common.Transition(
                    "initial_attempt_online",
                    self.offline,
                    self.attempt_online),  # 2
                secsgem.common.Transition(
                    "initial_host_offline",
                    self.offline,
                    self.host_offline),  # 2
                secsgem.common.Transition(
                    "switch_online",
                    self.equipment_offline,
                    self.attempt_online),  # 3
                secsgem.common.Transition(
                    "attempt_online_fail_equipment_offline",
                    self.attempt_online,
                    self.equipment_offline),  # 4
                secsgem.common.Transition(
                    "attempt_online_fail_host_offline",
                    self.attempt_online,
                    self.host_offline),  # 4
                secsgem.common.Transition(
                    "attempt_online_success",
                    self.attempt_online,
                    self.online),  # 5
                secsgem.common.Transition(
                    "switch_offline",
                    [self.online, self.online_local, self.online_remote],
                    self.equipment_offline),  # 6, 12
                secsgem.common.Transition(
                    "initial_online",
                    self.control,
                    self.online),  # 1
                secsgem.common.Transition(
                    "initial_online_local",
                    self.online,
                    self.online_local),  # 7
                secsgem.common.Transition(
                    "initial_online_remote",
                    self.online,
                    self.online_remote),  # 7
                secsgem.common.Transition(
                    "switch_online_local",
                    self.online_remote,
                    self.online_local),  # 8
                secsgem.common.Transition(
                    "switch_online_remote",
                    self.online_local,
                    self.online_remote),  # 9
                secsgem.common.Transition(
                    "remote_offline",
                    [self.online, self.online_local, self.online_remote],
                    self.host_offline),  # 10
                secsgem.common.Transition(
                    "remote_online",
                    self.host_offline,
                    self.online),  # 11
        ]

        # 1, forward online/offline depending on configuration
        self.control.events.enter.register(self._on_control_state_control)

        # 2, forward to configured offline state
        self.offline.events.enter.register(self._on_control_state_offline)

        # 7, forward to configured online state
        self.online.events.enter.register(self._on_control_state_online)

    def _on_control_state_control(self, _):
        if self._initial_control_state == "ONLINE":
            self._perform_transition("initial_online")
        else:
            self._perform_transition("initial_offline")

    def _on_control_state_offline(self, _):
        if self._initial_control_state == "EQUIPMENT_OFFLINE":
            self._perform_transition("initial_equipment_offline")
        elif self._initial_control_state == "ATTEMPT_ONLINE":
            self._perform_transition("initial_attempt_online")
        elif self._initial_control_state == "HOST_OFFLINE":
            self._perform_transition("initial_host_offline")

    def _on_control_state_online(self, _):
        if self._online_control_state == "REMOTE":
            self._perform_transition("initial_online_remote")
        else:
            self._perform_transition("initial_online_local")

    def start(self) -> None:
        """Perform start transition."""
        self._perform_transition("start")

    def attempt_online_fail_host_offline(self) -> None:
        """Perform attempt_online_fail_host_offline transition."""
        self._perform_transition("attempt_online_fail_host_offline")

    def attempt_online_success(self) -> None:
        """Perform attempt_online_success transition."""
        self._perform_transition("attempt_online_success")

    def switch_online(self) -> None:
        """Perform switch_online transition."""
        self._perform_transition("switch_online")

    def switch_offline(self) -> None:
        """Perform switch_offline transition."""
        self._perform_transition("switch_offline")

    def switch_online_local(self) -> None:
        """Perform switch_online_local transition."""
        self._perform_transition("switch_online_local")
        self._online_control_state = "LOCAL"

    def switch_online_remote(self) -> None:
        """Perform switch_online_remote transition."""
        self._perform_transition("switch_online_remote")
        self._online_control_state = "REMOTE"

    def remote_offline(self) -> None:
        """Perform remote_offline transition."""
        self._perform_transition("remote_offline")

    def remote_online(self) -> None:
        """Perform remote_online transition."""
        self._perform_transition("remote_online")



"""Contains helper functions."""
import errno
import sys
import types


def format_hex(text: bytes) -> str:
    """Return byte arrays (string) formated as hex numbers.

    Args:
        text: byte array

    Returns:
        formated sring

    Example:
        >>> import secsgem.common
        >>>
        >>> data = b"asdfg"
        >>> secsgem.common.format_hex(data)
        '61:73:64:66:67'

    """
    return ":".join(f"{c:02x}" for c in bytearray(text))


def is_windows() -> bool:
    """Return True if running on windows.

    Returns:
        True when running on a windows system

    """
    if sys.platform == "win32":
        return True

    return False


def function_name(function) -> str:
    """Get name of function or method.

    Returns:
        function/method name

    """
    if isinstance(function, types.FunctionType):
        return function.__name__

    return function.__self__.__class__.__name__ + "." + function.__name__


def indent_line(line: str, spaces: int = 2) -> str:
    """Indent line by a number of spaces.

    Args:
        line: input text
        spaces: number of spaces to prepend

    Returns:
        indented text

    """
    return f"{' ' * spaces}{line}"


def indent_block(block: str, spaces: int = 2) -> str:
    """Indent a multiline string by a number of spaces.

    Args:
        block: input text
        spaces: number of spaces to prepend to each line

    Returns:
        indented text

    """
    lines = block.split("\n")
    lines_filter = filter(None, lines)
    indented_lines = [indent_line(line, spaces) for line in lines_filter]
    return "\n".join(indented_lines)


def is_errorcode_ewouldblock(errorcode: int) -> bool:
    """Check if the errorcode is a would-block error.

    Args:
        errorcode: Code of the error

    Returns:
        True if blocking error code

    """
    if errorcode in (errno.EAGAIN, errno.EWOULDBLOCK):
        return True

    return False



"""Class for stream 02 function 41."""

from secsgem.secs.data_items import CPNAME, CPVAL, RCMD
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F41(SecsStreamFunction):
    """host command - send.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F41
        {
            RCMD: U1/I1/A
            PARAMS: [
                {
                    CPNAME: U1/U2/U4/U8/I1/I2/I4/I8/A
                    CPVAL: BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/A/B
                }
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F41({"RCMD": "COMMAND", "PARAMS": [{"CPNAME": "PARAM1", "CPVAL": "VAL1"},
        ...     {"CPNAME": "PARAM2", "CPVAL": "VAL2"}]})
        S2F41 W
          <L [2]
            <A "COMMAND">
            <L [2]
              <L [2]
                <A "PARAM1">
                <A "VAL1">
              >
              <L [2]
                <A "PARAM2">
                <A "VAL2">
              >
            >
          > .

    Data Items:
        - :class:`RCMD <secsgem.secs.data_items.RCMD>`
        - :class:`CPNAME <secsgem.secs.data_items.CPNAME>`
        - :class:`CPVAL <secsgem.secs.data_items.CPVAL>`

    """

    _stream = 2
    _function = 41

    _data_format = [
        RCMD,
        [
            [
                "PARAMS",
                CPNAME,
                CPVAL
            ]
        ]
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Class for stream 06 function 24."""

from secsgem.secs.data_items import RSDA
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS06F24(SecsStreamFunction):
    """Request spooled data acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F24
        RSDA: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F24(secsgem.secs.data_items.RSDA.ACK)
        S6F24
          <B 0x0> .

    Data Items:
        - :class:`RSDA <secsgem.secs.data_items.RSDA>`

    """

    _stream = 6
    _function = 24

    _data_format = RSDA

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 06 function 23."""

from secsgem.secs.data_items import RSDC
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS06F23(SecsStreamFunction):
    """Request spooled data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F23
        RSDC: U1[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F23(secsgem.secs.data_items.RSDC.PURGE)
        S6F23 W
          <U1 1 > .

    Data Items:
        - :class:`RSDC <secsgem.secs.data_items.RSDC>`

    """

    _stream = 6
    _function = 23

    _data_format = RSDC

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""SECS 8 byte float variable type."""

from .base_number import BaseNumber


class F8(BaseNumber):
    """Secs type for 8 byte float data.

    :param value: initial value
    :type value: list/float
    :param count: number of items this value
    :type count: integer
    """

    format_code = 0o40
    text_code = "F8"
    _base_type = float
    _min = -1.79769e+308
    _max = 1.79769e+308
    _bytes = 8
    _struct_code = "d"
    preferred_types = [float]



"""Class for stream 02 function 44."""

from secsgem.secs.data_items import FCNID, RSPACK, STRACK, STRID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F44(SecsStreamFunction):
    """reset spooling - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F44
        {
            RSPACK: B[1]
            DATA: [
                {
                    STRID: U1[1]
                    STRACK: B[1]
                    FCNID: [
                        DATA: U1[1]
                        ...
                    ]
                }
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F44({
        ...     "RSPACK": secsgem.secs.data_items.RSPACK.REJECTED,
        ...     "DATA": [
        ...         {"STRID": 1, "STRACK": secsgem.secs.data_items.STRACK.NOT_ALLOWED, "FCNID": [10]},
        ...         {"STRID": 2, "STRACK": secsgem.secs.data_items.STRACK.FUNCTION_UNKNOWN, "FCNID": [20]}]})
        S2F44
          <L [2]
            <B 0x1>
            <L [2]
              <L [3]
                <U1 1 >
                <B 0x1>
                <L [1]
                  <U1 10 >
                >
              >
              <L [3]
                <U1 2 >
                <B 0x3>
                <L [1]
                  <U1 20 >
                >
              >
            >
          > .

    Data Items:
        - :class:`RSPACK <secsgem.secs.data_items.RSPACK>`
        - :class:`STRID <secsgem.secs.data_items.STRID>`
        - :class:`STRACK <secsgem.secs.data_items.STRACK>`
        - :class:`FCNID <secsgem.secs.data_items.FCNID>`

    """

    _stream = 2
    _function = 44

    _data_format = [
        RSPACK,
        [
            [
                STRID,
                STRACK,
                [FCNID]
            ]
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 01 function 18."""

from secsgem.secs.data_items import ONLACK
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS01F18(SecsStreamFunction):
    """online acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F18
        ONLACK: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F18(secsgem.secs.data_items.ONLACK.ALREADY_ON)
        S1F18
          <B 0x2> .

    Data Items:
        - :class:`ONLACK <secsgem.secs.data_items.ONLACK>`

    """

    _stream = 1
    _function = 18

    _data_format = ONLACK

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Status Data Collection capability."""
from __future__ import annotations

import secsgem.secs

from .capability import Capability
from .handler import GemHandler
from .status_variable import StatusVariable, StatusVariableId


class StatusDataCollectionCapability(GemHandler, Capability):
    """Status Data Collection capability on GEM equipment."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize capability."""
        super().__init__(*args, **kwargs)

        self.__status_variables: dict[int | str, StatusVariable] = {
            StatusVariableId.CLOCK.value: StatusVariable(
                StatusVariableId.CLOCK,
                "Clock",
                "",
                secsgem.secs.variables.String),
            StatusVariableId.CONTROL_STATE.value: StatusVariable(
                StatusVariableId.CONTROL_STATE,
                "ControlState",
                "",
                secsgem.secs.variables.Binary),
            StatusVariableId.EVENTS_ENABLED.value: StatusVariable(
                StatusVariableId.EVENTS_ENABLED,
                "EventsEnabled",
                "",
                secsgem.secs.variables.Array),
            StatusVariableId.ALARMS_ENABLED.value: StatusVariable(
                StatusVariableId.ALARMS_ENABLED,
                "AlarmsEnabled",
                "",
                secsgem.secs.variables.Array),
            StatusVariableId.ALARMS_SET.value: StatusVariable(
                StatusVariableId.ALARMS_SET,
                "AlarmsSet",
                "",
                secsgem.secs.variables.Array),
        }

    @property
    def _status_variables(self) -> dict[int | str, StatusVariable]:
        return self.__status_variables

    @property
    def status_variables(self) -> dict[int | str, StatusVariable]:
        """Get list of the status variables.

        Returns:
            Status variable list

        """
        return self._status_variables

    def on_sv_value_request(self,
                            svid: secsgem.secs.variables.Base,
                            status_variable: StatusVariable) -> secsgem.secs.variables.Base:
        """Get the status variable value depending on its configuation.

        Override in inherited class to provide custom status variable request handling.

        Args:
            svid: Id of the status variable encoded in the corresponding type
            status_variable: The status variable requested

        Returns:
            The value encoded in the corresponding type

        """
        del svid  # unused variable

        return status_variable.value_type(status_variable.value)

    def _get_sv_value(self, status_variable: StatusVariable) -> secsgem.secs.variables.Base:
        """Get the status variable value depending on its configuation.

        Args:
            status_variable: The status variable requested

        Returns:
            The value encoded in the corresponding type

        """
        if status_variable.svid == StatusVariableId.CLOCK.value:
            result = status_variable.value_type(self._get_clock())
        elif status_variable.svid == StatusVariableId.CONTROL_STATE.value:
            result = status_variable.value_type(self._get_control_state_id())
        elif status_variable.svid == StatusVariableId.EVENTS_ENABLED.value:
            events = self._get_events_enabled()
            result = status_variable.value_type(secsgem.secs.data_items.SV, events)
        elif status_variable.svid == StatusVariableId.ALARMS_ENABLED.value:
            alarms = self._get_alarms_enabled()
            result = status_variable.value_type(secsgem.secs.data_items.SV, alarms)
        elif status_variable.svid == StatusVariableId.ALARMS_SET.value:
            alarms = self._get_alarms_set()
            result = status_variable.value_type(secsgem.secs.data_items.SV, alarms)
        else:
            if status_variable.use_callback:
                result = self.on_sv_value_request(status_variable.id_type(status_variable.svid), status_variable)
            else:
                result = status_variable.value_type(status_variable.value)

        return result

    def _on_s01f03(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 1, Function 3, Equipment status request.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler  # unused parameters

        function = self.settings.streams_functions.decode(message)

        responses = []

        if len(function) == 0:
            responses = [self._get_sv_value(status_variable) for status_variable in self._status_variables.values()]
        else:
            for status_variable_id in function:
                if status_variable_id not in self._status_variables:
                    responses.append(secsgem.secs.variables.Array(secsgem.secs.data_items.SV, []))
                else:
                    status_variable = self._status_variables[status_variable_id]
                    responses.append(self._get_sv_value(status_variable))

        return self.stream_function(1, 4)(responses)

    def _on_s01f11(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 1, Function 11, SV namelist request.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler  # unused parameters

        function = self.settings.streams_functions.decode(message)

        responses = []

        if len(function) == 0:
            responses = [{
                "SVID": status_variable.svid,
                "SVNAME": status_variable.name,
                "UNITS": status_variable.unit,
            } for status_variable in self._status_variables.values()]
        else:
            for status_variable_id in function:
                if status_variable_id not in self._status_variables:
                    responses.append({"SVID": status_variable_id, "SVNAME": "", "UNITS": ""})
                else:
                    status_variable = self._status_variables[status_variable_id]
                    responses.append({"SVID": status_variable.svid,
                                      "SVNAME": status_variable.name,
                                      "UNITS": status_variable.unit})

        return self.stream_function(1, 12)(responses)



"""IDTYP data item."""
from .. import variables
from .base import DataItemBase


class IDTYP(DataItemBase):
    """ID type.

    :Type: :class:`Binary <secsgem.secs.variables.Binary>`
    :Length: 1

    **Values**
        +-------+-------------------+-------------------------------------------------------+
        | Value | Description       | Constant                                              |
        +=======+===================+=======================================================+
        | 0     | Wafer ID          | :const:`secsgem.secs.data_items.IDTYP.WAFER`          |
        +-------+-------------------+-------------------------------------------------------+
        | 1     | Wafer cassette ID | :const:`secsgem.secs.data_items.IDTYP.WAFER_CASSETTE` |
        +-------+-------------------+-------------------------------------------------------+
        | 2     | Film frame ID     | :const:`secsgem.secs.data_items.IDTYP.FILM_FRAME`     |
        +-------+-------------------+-------------------------------------------------------+
        | 3-63  | Reserved, error   |                                                       |
        +-------+-------------------+-------------------------------------------------------+

    **Used In Function**
        - :class:`SecsS12F01 <secsgem.secs.functions.SecsS12F01>`
        - :class:`SecsS12F03 <secsgem.secs.functions.SecsS12F03>`
        - :class:`SecsS12F04 <secsgem.secs.functions.SecsS12F04>`
        - :class:`SecsS12F05 <secsgem.secs.functions.SecsS12F05>`
        - :class:`SecsS12F07 <secsgem.secs.functions.SecsS12F07>`
        - :class:`SecsS12F09 <secsgem.secs.functions.SecsS12F09>`
        - :class:`SecsS12F11 <secsgem.secs.functions.SecsS12F11>`
        - :class:`SecsS12F13 <secsgem.secs.functions.SecsS12F13>`
        - :class:`SecsS12F14 <secsgem.secs.functions.SecsS12F14>`
        - :class:`SecsS12F15 <secsgem.secs.functions.SecsS12F15>`
        - :class:`SecsS12F16 <secsgem.secs.functions.SecsS12F16>`
        - :class:`SecsS12F17 <secsgem.secs.functions.SecsS12F17>`
        - :class:`SecsS12F18 <secsgem.secs.functions.SecsS12F18>`

    """

    __type__ = variables.Binary
    __count__ = 1

    WAFER = 0
    WAFER_CASSETTE = 1
    FILM_FRAME = 2



"""Class for stream 02 function 16."""

from secsgem.secs.data_items import EAC
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F16(SecsStreamFunction):
    """new equipment constant - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F16
        EAC: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F16(secsgem.secs.data_items.EAC.BUSY)
        S2F16
          <B 0x2> .

    Data Items:
        - :class:`EAC <secsgem.secs.data_items.EAC>`

    """

    _stream = 2
    _function = 16

    _data_format = EAC

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 05 function 04."""

from secsgem.secs.data_items import ACKC5
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS05F04(SecsStreamFunction):
    """en-/disable alarm - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F04
        ACKC5: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F04(secsgem.secs.data_items.ACKC5.ACCEPTED)
        S5F4
          <B 0x0> .

    Data Items:
        - :class:`ACKC5 <secsgem.secs.data_items.ACKC5>`

    """

    _stream = 5
    _function = 4

    _data_format = ACKC5

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""State machine for communication state."""
from __future__ import annotations

import enum
import threading

import secsgem.common


class CommunicationState(enum.Enum):
    """States for connection state machine."""

    DISABLED = 0
    ENABLED = 1
    NOT_COMMUNICATING = 2
    HOST_INITIATED_CONNECT = 3
    WAIT_CR_FROM_HOST = 4
    EQUIPMENT_INITIATED_CONNECT = 5
    WAIT_DELAY = 6
    WAIT_CRA = 7
    COMMUNICATING = 8


class CommunicationStateMachine(secsgem.common.StateMachine):  # pylint: disable=too-many-instance-attributes
    """Communication state machine.

    Not going to HOST_INITIATED_CONNECT because two states at a time are not supported.
    There is a transistion to get out of EQUIPMENT_INITIATED_CONNECT when the HOST_INITIATED_CONNECT happens.
    """

    def __init__(self, settings: secsgem.common.Settings) -> None:
        """Initialize state machine.

        Args:
            settings: communication settings

        """
        super().__init__()

        self._settings = settings

        self.disabled = secsgem.common.State(
            CommunicationState.DISABLED,
            "DISABLED",
            initial=True)
        self.enabled = secsgem.common.State(
            CommunicationState.ENABLED,
            "ENABLED")
        self.not_communicating = secsgem.common.State(
            CommunicationState.NOT_COMMUNICATING,
            "NOT_COMMUNICATING",
            parent=self.enabled)
        self.host_initiated_connect = secsgem.common.State(
            CommunicationState.HOST_INITIATED_CONNECT,
            "HOST_INITIATED_CONNECT",
            parent=self.enabled)
        self.wait_cr_from_host = secsgem.common.State(
            CommunicationState.WAIT_CR_FROM_HOST,
            "WAIT_CR_FROM_HOST",
            parent=self.enabled)
        self.equipment_initiated_connect = secsgem.common.State(
            CommunicationState.EQUIPMENT_INITIATED_CONNECT,
            "EQUIPMENT_INITIATED_CONNECT",
            parent=self.enabled)
        self.wait_delay = secsgem.common.State(
            CommunicationState.WAIT_DELAY,
            "WAIT_DELAY",
            parent=self.enabled)
        self.wait_cra = secsgem.common.State(
            CommunicationState.WAIT_CRA,
            "WAIT_CRA",
            parent=self.enabled)
        self.communicating = secsgem.common.State(
            CommunicationState.COMMUNICATING,
            "COMMUNICATING",
            parent=self.enabled)

        # transition 1
        self._current_state: secsgem.common.State = self.disabled

        self._transitions: list[secsgem.common.Transition] = [
            secsgem.common.Transition("enable", self.disabled, self.not_communicating),  # 2 and 4
            secsgem.common.Transition("disable",
                                      [
                    self.enabled,
                    self.not_communicating,
                    self.communicating,
                    self.equipment_initiated_connect,
                    self.wait_delay,
                    self.wait_cra,
                    self.host_initiated_connect,
                    self.wait_cr_from_host,
                ],
                                      self.disabled),  # 3
            secsgem.common.Transition("select", [self.not_communicating, self.communicating],
                                      self.wait_cra),  # 5
            secsgem.common.Transition("communicationreqfail", self.wait_cra, self.wait_delay),  # 6
            secsgem.common.Transition("delayexpired", self.wait_delay, self.wait_cra),  # 7
            secsgem.common.Transition("messagereceived", self.wait_delay, self.wait_cra),  # 8
            secsgem.common.Transition("s1f14received", self.wait_cra, self.communicating),  # 9
            secsgem.common.Transition("communicationfail", self.communicating, self.not_communicating),  # 14
            secsgem.common.Transition("s1f13received",
                                      [
                    self.wait_cr_from_host,
                    self.wait_delay,
                    self.wait_cra,
                ],
                                      self.communicating),  # 14
        ]

        self._wait_cra_timer: threading.Thread | None = None
        self._comm_delay_timer: threading.Thread | None = None

        self.wait_cra.events.enter.register(self._on_state_wait_cra)
        self.wait_delay.events.enter.register(self._on_state_wait_delay)
        self.wait_cra.events.leave.register(self._on_state_leave_wait_cra)
        self.wait_delay.events.leave.register(self._on_state_leave_wait_delay)

    def _on_state_wait_cra(self, _):
        """Connection state model changed to state WAIT_CRA.

        Args:
            data: event attributes

        """
        self._wait_cra_timer = threading.Timer(self._settings.timeouts.t3, self._on_wait_cra_timeout)
        self._wait_cra_timer.start()

    def _on_state_wait_delay(self, _):
        """Connection state model changed to state WAIT_DELAY.

        Args:
            data: event attributes

        """
        self._comm_delay_timer = threading.Timer(self._settings.establish_communication_timeout,
                                                 self._on_wait_comm_delay_timeout)
        self._comm_delay_timer.start()

    def _on_state_leave_wait_cra(self, _):
        """Connection state model changed to state WAIT_CRA.

        Args:
            data: event attributes

        """
        if self._wait_cra_timer is not None:
            self._wait_cra_timer.cancel()

    def _on_state_leave_wait_delay(self, _):
        """Connection state model changed to state WAIT_DELAY.

        Args:
            data: event attributes

        """
        if self._comm_delay_timer is not None:
            self._comm_delay_timer.cancel()

    def _on_wait_cra_timeout(self):
        """Linktest time timed out, so send linktest request."""
        self._perform_transition("communicationreqfail")

    def _on_wait_comm_delay_timeout(self):
        """Linktest time timed out, so send linktest request."""
        self._perform_transition("delayexpired")

    def enable(self) -> None:
        """Perform enable transition."""
        self._perform_transition("enable")

    def disable(self) -> None:
        """Perform disable transition."""
        self._perform_transition("disable")

    def select(self) -> None:
        """Perform select transition."""
        self._perform_transition("select")

    def messagereceived(self) -> None:
        """Perform messagereceived transition."""
        self._perform_transition("messagereceived")

    def s1f14received(self) -> None:
        """Perform s1f14received transition."""
        self._perform_transition("s1f14received")

    def communicationfail(self) -> None:
        """Perform communicationfail transition."""
        self._perform_transition("communicationfail")

    def s1f13received(self) -> None:
        """Perform s1f13received transition."""
        self._perform_transition("s1f13received")



"""OBJSPEC data item."""
from .. import variables
from .base import DataItemBase


class OBJSPEC(DataItemBase):
    """Specific object instance.

    :Type: :class:`String <secsgem.secs.variables.String>`

    **Used In Function**
        - :class:`SecsS02F49 <secsgem.secs.functions.SecsS02F49>`
        - :class:`SecsS14F01 <secsgem.secs.functions.SecsS14F01>`
        - :class:`SecsS14F03 <secsgem.secs.functions.SecsS14F03>`

    """

    __type__ = variables.String



"""Header for the hsms linktest response."""

from .header import HsmsHeader, HsmsSType


class HsmsLinktestRspHeader(HsmsHeader):
    """Header for Linktest Response.

    Header for message with SType 6.
    """

    def __init__(self, system: int):
        """Initialize a hsms linktest response.

        Args:
            system: message ID

        Example:
            >>> import secsgem.hsms
            >>>
            >>> secsgem.hsms.HsmsLinktestRspHeader(10)
            HsmsLinktestRspHeader({session_id:0xffff, stream:00, function:00, p_type:0x00, s_type:0x06, \
system:0x0000000a, require_response:False})

        """
        super().__init__(system, 0xFFFF, 0, 0, False, 0x00, HsmsSType.LINKTEST_RSP)



"""Class for stream 07 function 26."""

from secsgem.secs.data_items import PPID, MDLN, SOFTREV,CCODE, PPARM
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS07F26(SecsStreamFunction):
    """Format process program data..

    Args:
        value: parameters for this function (see example)
    """

    _stream = 7
    _function = 26

    _data_format = [
        PPID,
        MDLN,
        SOFTREV,
        [
            [
                "CCODE",
                CCODE,
                [PPARM]
            ]
        ]
    ]

    _to_host = True
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = True
